#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <libtasn1.h>

const asn1_static_node basic_asn1_tab[] = {
  { "BASIC", 536872976, NULL },
  { NULL, 1073741836, NULL },
  { "Any", 1073741837, NULL },
  { "ObjectIdentifier", 12, NULL },
  { NULL, 0, NULL }
};
